from tweepy import __version__ as _tweepy_ver

_TWEEPY_VER_TUPLE: tuple[int] = tuple(int(n) for n in _tweepy_ver.split(".", 3))


def tweepy_ver_before(required_version: str):
    major, minor, patch = (int(n) for n in required_version.split(".", 3))
    return _TWEEPY_VER_TUPLE[0] < major or (
        _TWEEPY_VER_TUPLE[0] == major
        and (
            _TWEEPY_VER_TUPLE[1] < minor
            or (_TWEEPY_VER_TUPLE[1] == minor and _TWEEPY_VER_TUPLE[2] < patch)
        )
    )



import asyncio
import datetime
import json
import logging
from os import path
from typing import Any, NamedTuple, Optional

from pymitter import EventEmitter
from sse_starlette import ServerSentEvent

from .settings import Settings, get_settings
from snowflake import TWEPOCH, make_snowflake

SSE_EVENT_NAME = "announce-sse"

_logger = logging.getLogger(__name__)


# https://stackoverflow.com/a/1363857/11524425
class SingletonEventEmitter(EventEmitter):
    __instance = None

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance


class SentEvent(NamedTuple):
    timestamp: datetime.datetime
    event: ServerSentEvent


def make_sse(
    data: Optional[Any] = None,
    *,
    id: Optional[int] = None,
    event: Optional[str] = None,
    retry: Optional[int] = None,
    comment: Optional[str] = None,
):
    r"""
    Create a `ServerSentEvent` with `\n` (LF) line endings

    Parameters
    ----------
    data : Any | None
        The data field for the message. Any datatype can be specified to this parameter,
        since the SSE instance converts them to a string before they are sent.
    id: int | None
        The event ID to set the EventSource object's last event ID value to.
    event: str | None
        The event's type. If this is specified, an event will be dispatched on the
        client to the listener for the specified event name; a website would use
        `addEventListener()` to listen for named events. The default event type
        is "message".
    retry: int | None
        The reconnection time to use when attempting to send the event. This must be an
        integer, specifying the reconnection time in milliseconds. If a non-integer
        value is specified, the field is ignored.
    comment: str | None
        A colon as the first character of a line is essence a comment, and is ignored.
        Usually used as a ping message to keep connecting. If set, this will be a
        comment message.
    """
    return ServerSentEvent(
        data, id=id, event=event, retry=retry, comment=comment, sep="\n"
    )


# Based on https://maxhalford.github.io/blog/flask-sse-no-deps/ (MIT License)
class MessageAnnouncer:
    _ee = SingletonEventEmitter()

    def __init__(self, max_event_age: int = 15, epoch: int = TWEPOCH):
        self.listeners: list[asyncio.Queue] = []
        self.recent_events: list[SentEvent] = []
        self.max_event_age = max_event_age
        self._id_epoch = epoch

        self._ee.on("announce-sse", self.announce)

        settings: Settings = get_settings()
        self.recent_events_file: str = settings.recent_events_file

        # Load recent events from file
        if path.exists(self.recent_events_file):
            with open(self.recent_events_file) as f:
                try:
                    data: list[dict] = json.load(f)
                    for event in data:
                        sse = ServerSentEvent(**event["sse_data"])
                        timestamp_dt = datetime.datetime.fromtimestamp(
                            event["timestamp"], datetime.timezone.utc
                        )
                        self.recent_events.append(SentEvent(timestamp_dt, sse))
                except Exception:  # noqa: PIE786
                    _logger.exception("Failed to read %s", self.recent_events_file)
        event_count = len(self.recent_events)
        if event_count > 0:
            _logger.debug(
                "Loaded %s recently-sent event%s from %s",
                event_count,
                "s" if event_count != 1 else "",
                self.recent_events_file,
            )

    def listen(self) -> asyncio.Queue:
        """
        Subscribe to events sent by the `MessageAnnouncer` instance
        """
        # maxsize param defines maximum number of events that can be sent at once
        msg_q = asyncio.Queue(maxsize=10)
        self.listeners.append(msg_q)
        return msg_q

    def save_recent_events(self):
        """
        Saves a list of recently-sent server-side events to a file for later use.

        The path name of the target file is specified in your application's settings file.
        """

        event_dict_list_temp: list[dict] = []
        for event in self.recent_events:
            timestamp, sse = event
            event_dict_list_temp.append(
                {
                    "timestamp": timestamp.timestamp(),
                    "sse_data": {
                        "id": sse.id,
                        "event": sse.event,
                        "data": sse.data,
                        "retry": sse.retry,
                        "comment": sse.comment,
                    },
                }
            )

        with open(self.recent_events_file, "w") as f:
            f.write(json.dumps(event_dict_list_temp))
            _logger.debug("Recent events saved to %s", self.recent_events_file)

    def announce(self, sse: ServerSentEvent):
        """
        Push a new server-sent event to all subscribed clients

        Parameters
        ----------
        sse : ServerSentEvent
            The server-sent event to send
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        sse.id = make_snowflake(int(now.timestamp()), 1, 1, 1, self._id_epoch)

        for i in range(len(self.listeners) - 1, -1, -1):
            try:
                self.listeners[i].put_nowait(sse)
            except asyncio.QueueFull:
                del self.listeners[i]
        _logger.debug(
            "Pushed SSE data:\n%s",
            {
                "id": sse.id,
                "event": sse.event,
                "data": sse.data,
                "retry": sse.retry,
                "comment": sse.comment,
            },
        )
        self.recent_events.append(SentEvent(now, sse))
        self.save_recent_events()

    def clean_up_recent_events(self):
        """
        Remove recently-sent server-side events from history that are older than the
        number of minutes specified in `max_event_age`.

        By default, events from more than 15 minutes ago will be purged.
        """
        dt = datetime.datetime.now(datetime.timezone.utc)
        self.recent_events = list(
            filter(
                lambda e: dt - e.timestamp
                <= datetime.timedelta(minutes=self.max_event_age),
                self.recent_events,
            )
        )

    def get_recent_events(self, last_id: int) -> list[SentEvent]:
        """
        Retrieve the most recently-sent events, oldest first, from the given ID onwards

        Parameters
        ----------
        last_id : int
            The snowflake ID of the last received tweet
        """
        self.clean_up_recent_events()
        return [
            e.event for e in filter(lambda e: last_id < e.event.id, self.recent_events)
        ]



import asyncio
import json
import logging
from ipaddress import ip_address
from typing import (
    Any,
    AsyncContextManager,
    Awaitable,
    Callable,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)

from sse_starlette.sse import AppStatus, EventSourceResponse, ServerSentEvent
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route

from .event import MessageAnnouncer, make_sse
from .processor import BaseProcessor
from .settings import Settings, get_settings
from .stream import StreamClient
from snowflake import TWEPOCH

_logger = logging.getLogger(__name__)


class TweetRelay(Starlette):
    """
    This class manages dependencies between the Twitter API stream client (which
    listens for new tweets in realtime), stream response processors (which processes a
    tweet every time one is received), and the message announcer (which is responsible
    for sending server-sent events to all connected clients)
    """

    stream_client: StreamClient

    def __init__(
        self,
        processors: Sequence[BaseProcessor],
        expansions_fields: Optional[dict] = None,
        settings: Optional[Settings] = None,
        stream_client_cls: Type = StreamClient,
        sse_id_epoch: int = TWEPOCH,
        max_event_age: int = 15,
        debug: bool = False,
        middleware: Optional[Sequence[Middleware]] = None,
        exception_handlers: Optional[
            Mapping[
                Any,
                Callable[
                    [Request, Exception],
                    Union[Response, Awaitable[Response]],
                ],
            ]
        ] = None,
        lifespan: Optional[Callable[["Starlette"], AsyncContextManager]] = None,
    ):
        routes = [Route("/", self.stream)]
        on_startup = [self.startup]
        on_shutdown = [self.shutdown]
        super().__init__(
            debug,
            routes,
            middleware,
            exception_handlers,
            on_startup,
            on_shutdown,
            lifespan,
        )

        self.announcer = MessageAnnouncer(max_event_age, sse_id_epoch)

        if settings is None:
            settings = get_settings()
        assert isinstance(stream_client_cls, type) and issubclass(
            stream_client_cls, StreamClient
        ), "stream_client_cls must be of type StreamClient or a subclass of it"
        self.stream_client = stream_client_cls(
            settings.bearer_token, processors, expansions_fields or {}
        )

    def startup(self):
        """
        Connects the stream client when server starts up
        """
        if not self.stream_client.running:
            self.stream_client.run_forever()

    async def shutdown(self):
        """
        Disconnects stream client when server shuts down
        """
        if (
            isinstance(self.stream_client, StreamClient)
            and not self.stream_client.task.cancelled()
        ):
            self.stream_client.disconnect()
            # App shutdown is not over until `stream_client.task` finishes
            await self.stream_client.task

    @staticmethod
    async def event_generator(announcer: MessageAnnouncer, request: Request):
        sse_queue = announcer.listen()
        ip = ip_address(request.client.host)
        client_addr = f"{ip.compressed}"
        if request.client.port:
            client_addr = f"{client_addr}:{request.client.port}"

        # Send event backfill if Last-Event-ID header was provided
        last_id = int(request.headers.get("Last-Event-ID", "0"))
        if last_id:
            _logger.debug("Received Last-Event-ID: %s", last_id)
            for event in announcer.get_recent_events(last_id):
                yield event

        try:
            while True:
                # Wait for new events to arrive and send them to all connected clients
                # as they come in
                msg: ServerSentEvent = await sse_queue.get()
                yield msg
        except asyncio.CancelledError:
            # Clean up listener on disconnect/server shutdown
            try:
                announcer.listeners.remove(sse_queue)
            except ValueError:
                pass
            _logger.debug(
                "Cleaned up message queue for aborted connection from %s",
                client_addr,
            )

        if AppStatus.should_exit:
            # App is shutting down/reloading
            _logger.debug("Sending disconnect event to %s", client_addr)
            yield make_sse(
                json.dumps({"message": "Server is shutting down or restarting"}),
                event="disconnect",
                retry=25_000,
            )
        else:
            # Check if connection was really disconnected
            if await request.is_disconnected():
                _logger.debug("Connection from %s was disconnected", client_addr)

    async def stream(self, request: Request):
        """
        The main endpoint for pushing server-sent events to a client

        Parameters
        ----------
        request: Request
            An incoming request from a connected client
        """
        return EventSourceResponse(
            self.event_generator(self.announcer, request), ping=20
        )



"""
snowflake.py by Falcon Dai (https://github.com/falcondai/python-snowflake), MIT License

Modified for Python 3 with some additions and style fixes
"""

from datetime import datetime, timezone

# Twitter's snowflake parameters
TWEPOCH = 1288834974657
DATACENTER_ID_BITS = 5
WORKER_ID_BITS = 5
SEQUENCE_ID_BITS = 12
MAX_DATACENTER_ID = 1 << DATACENTER_ID_BITS
MAX_WORKER_ID = 1 << WORKER_ID_BITS
MAX_SEQUENCE_ID = 1 << SEQUENCE_ID_BITS
MAX_TIMESTAMP = 1 << (64 - DATACENTER_ID_BITS - WORKER_ID_BITS - SEQUENCE_ID_BITS)


def make_snowflake(
    timestamp_ms: int,
    datacenter_id: int,
    worker_id: int,
    sequence_id: int,
    epoch: int = TWEPOCH,
) -> int:
    """
    Generate a Twitter Snowflake ID, based on
    https://github.com/twitter-archive/snowflake/blob/snowflake-2010/src/main/scala/com/twitter/service/snowflake/IdWorker.scala

    Parameters
    ----------
    timestamp_ms : int
        Time since Unix epoch in milliseconds

    epoch : int
        Unix timestamp of starting epoch in milliseconds.
        Defaults to `snowflake.TWEPOCH` (1288834974657)
    """

    sid = (
        ((int(timestamp_ms) - epoch) % MAX_TIMESTAMP)
        << DATACENTER_ID_BITS
        << WORKER_ID_BITS
        << SEQUENCE_ID_BITS
    )
    sid += (datacenter_id % MAX_DATACENTER_ID) << WORKER_ID_BITS << SEQUENCE_ID_BITS
    sid += (worker_id % MAX_WORKER_ID) << SEQUENCE_ID_BITS
    sid += sequence_id % MAX_SEQUENCE_ID
    return sid


def melt(snowflake_id: int, epoch: int = TWEPOCH) -> tuple:
    """
    Inversely transform a Snowflake ID back to its parts.
    """
    sequence_id = snowflake_id & (MAX_SEQUENCE_ID - 1)
    worker_id = (snowflake_id >> SEQUENCE_ID_BITS) & (MAX_WORKER_ID - 1)
    datacenter_id = (snowflake_id >> SEQUENCE_ID_BITS >> WORKER_ID_BITS) & (
        MAX_DATACENTER_ID - 1
    )
    timestamp_ms = (
        snowflake_id >> SEQUENCE_ID_BITS >> WORKER_ID_BITS >> DATACENTER_ID_BITS
    )
    timestamp_ms += epoch

    return (timestamp_ms, int(datacenter_id), int(worker_id), int(sequence_id))


def to_datetime(timestamp_ms: int, localtime: bool = False) -> datetime:
    """
    Convert a timestamp with milliseconds to a datetime object.

    Parameters
    ----------
    localtime : bool
        Whether to return as a `datetime` object in the machine's local timezone or
        UTC (default)
    """
    if localtime:
        return datetime.fromtimestamp(timestamp_ms // 1000).astimezone()
    return datetime.fromtimestamp(timestamp_ms // 1000, timezone.utc)



from functools import lru_cache

from pydantic import BaseSettings, validator


class Settings(BaseSettings):
    """
    Reads application settings from a file and stores them into an encapsulated object.
    Currently, only reading from files named `.env` in the current working directory
    is supported.
    """

    bearer_token: str
    host: str = "127.0.0.1"
    port: int = 8000
    logfile_level: str = "DEBUG"
    recent_events_file: str = "recent_events.json"

    @validator("logfile_level")
    def log_level_is_valid(cls, value: str):
        value = value.strip().upper()
        if value not in ("TRACE", "DEBUG", "INFO", "WARNING", "ERROR"):
            raise ValueError(
                "Value must be either one of the following: "
                "TRACE, DEBUG, INFO, WARNING, ERROR"
            )
        return value

    class Config:
        env_file = ".env"


@lru_cache
def get_settings():
    return Settings()  # type: ignore



from tweepy import StreamResponse  # noqa: TC002

from .event import SSE_EVENT_NAME, SingletonEventEmitter, make_sse


class BaseProcessor:
    """
    The base class for creating stream response (tweet) processors, which perform
    specific actions every time it receives a tweet from the stream client.
    """

    _ee = SingletonEventEmitter()

    async def on_data(self, data: StreamResponse):
        """
        |coroutine|

        This is called when a tweet was received from the stream client.

        Parameters
        ----------
        data : StreamResponse
            A response (tweet or error message) from Twitter that was picked
            up by the stream client
        """
        raise NotImplementedError

    async def announce(self, sse_event_name: str, payload: str):
        """
        Create a new `ServerSentEvent` object and send it to the message announcer

        Parameters
        ----------
        sse_event_name : str
            The name of the event type
        payload : str
            A string of data to be sent
        """
        sse = make_sse(event=sse_event_name, data=payload)
        await self._ee.emit_async(SSE_EVENT_NAME, sse)



import asyncio
import logging
import sys
import time
from typing import Iterable, Sequence

import aiohttp
from tweepy import StreamResponse
from tweepy.asynchronous.streaming import AsyncStreamingClient

from ._utils import tweepy_ver_before
from .event import SingletonEventEmitter
from .processor import BaseProcessor

TRACE_LOG_LEVEL = 5
_logger = logging.getLogger(__name__)


class StreamClient(AsyncStreamingClient):
    """
    A Twitter API v2 filtered stream client that listens for new tweets as soon as they
    are published and passes those tweets to stream response processors
    """

    _ee = SingletonEventEmitter()
    running = False

    def __init__(
        self,
        bearer_token: str,
        processors: Sequence[BaseProcessor],
        expansions_fields: dict[str, Iterable[str]],
        *args,
        **kwargs,
    ):
        self._expansions_fields = expansions_fields
        for processor in processors:
            self._ee.on("on_data", processor.on_data)

        return super().__init__(bearer_token, *args, **kwargs)

    async def on_connect(self):
        return await super().on_connect()

    async def on_connection_error(self):
        msg = "Stream connection has errored or timed out. Reconnecting the stream"
        if tweepy_ver_before("4.12.1"):
            exc_type, exc_value, _trace = sys.exc_info()
            _logger.error("%s\n%s: %s", msg, str(exc_type)[8:-2], exc_value)
        else:
            _logger.error(msg)

    async def on_closed(self, response: aiohttp.ClientResponse):
        if not self.task.cancelled():
            _logger.error(
                "Stream connection closed by Twitter. Reconnecting the stream"
            )

    async def on_keep_alive(self):
        _logger.log(TRACE_LOG_LEVEL, "Received keep-alive signal")
        self.last_keep_alive_time = time.time()

    async def on_request_error(self, status_code):
        _logger.error(
            "Stream encountered HTTP Error: %d. Reconnecting the stream", status_code
        )

    async def on_response(self, response: StreamResponse):
        await super().on_response(response)
        # Emit events to processors only when a tweet was received
        if response.data:
            await self._ee.emit_async("on_data", response)

    def disconnect(self):
        self.running = False
        if not self.session.closed:
            _logger.info("Disconnecting the stream")
        return super().disconnect()

    def run(self) -> asyncio.Task:
        self.running = True
        _logger.info("Connecting the stream")
        return self.filter(**self._expansions_fields)

    def run_forever(self):
        """
        Run the stream forever (even after uncaught exceptions) until we manually
        call `disconnect()`
        """
        self.running = True

        async def task():
            while self.running:
                await self.run()
                if self.running:
                    _logger.info("Restarting the stream")
                    await asyncio.sleep(1)

        return asyncio.create_task(task())
